%
%
%  Vesicle Outside Cell Extension for Imaris 7.3.0
%
%  Copyright Bitplane AG 2011
%
%
%  Installation:
%
%  - Copy this file into the XTensions folder in the Imaris installation directory
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%        <Item name="Vesicle Outside Cell" icon="Matlab" tooltip="For every Spots Compute the Distance to the nearest Cell and the Cell ID.">
%          <Command>MatlabXT::XTVesicleOutsideCell(%i)</Command>
%        </Item>
%      </Menu>
%      <SurpassTab>
%        <SurpassComponent name="bpCells">
%        <Item name="Vesicle Outside Cell" icon="Matlab" tooltip="For every Spots Compute the Distance to the nearest Cell and the Cell ID.">
%          <Command>MatlabXT::XTVesicleOutsideCell(%i)</Command>
%        </Item>
%        </SurpassComponent>
%        <SurpassComponent name="bpSpots">
%        <Item name="Vesicle Outside Cell" icon="Matlab" tooltip="For every Spots Compute the Distance to the nearest Cell and the Cell ID.">
%          <Command>MatlabXT::XTVesicleOutsideCell(%i)</Command>
%        </Item>
%        </SurpassComponent>
%      </SurpassTab>
%    </CustomTools>
%  
%
%  Description:
%
% For one Cells and Spots Surpass Component, find for every Spot the
% nearest Cell. Add the Distance and Cell ID as Statistics Values to the
% spots.
%
%
function XTVesicleOutsideCell(aImarisApplicationID)
if isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
    vImarisApplication = aImarisApplicationID;
else
    % connect to Imaris interface
    javaaddpath ImarisLib.jar
    vImarisLib = ImarisLib;
    if ischar(aImarisApplicationID)
        aImarisApplicationID = round(str2double(aImarisApplicationID));
    end
    vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
end

% the user has to create a scene with some spots and surface
vSurpassScene = vImarisApplication.GetSurpassScene;
if isequal(vSurpassScene, [])
    msgbox('Please create some Spots and Cells in the Surpass Scene')
    return
end

% get the spots and the surface object
vSpots = vImarisApplication.GetFactory.ToSpots(vImarisApplication.GetSurpassSelection);
vCells = vImarisApplication.GetFactory.ToCells(vImarisApplication.GetSurpassSelection);

vSpotsSelected = ~isequal(vSpots, []);
vCellSelected = ~isequal(vCells, []);
if vSpotsSelected
    vParent = vSpots.GetParent;
elseif vCellSelected
    vParent = vCells.GetParent;
else
    vParent = vSurpassScene;
end

% prompt if there is more than one spots or cells object
vSpotsSelection = 1;
vCellsSelection = 1;
vNumberOfSpots = 0;
vNumberOfCells = 0;
vSpotsList = [];
vCellsList = [];
vSpotsName = {};
vCellsName = {};
for vIndex = 1:vParent.GetNumberOfChildren
    vItem = vParent.GetChild(vIndex-1);
    if vImarisApplication.GetFactory.IsSpots(vItem)
        vNumberOfSpots = vNumberOfSpots + 1;
        vSpotsList(vNumberOfSpots) = vIndex;
        vSpotsName{vNumberOfSpots} = char(vItem.GetName);
        
        if vSpotsSelected && strcmp(vItem.GetName, vSpots.GetName)
            vSpotsSelection = vNumberOfSpots; 
        end
    elseif vImarisApplication.GetFactory.IsCells(vItem)
        vNumberOfCells = vNumberOfCells + 1;
        vCellsList(vNumberOfCells) = vIndex;
        vCellsName{vNumberOfCells} = char(vItem.GetName);
        
        if vCellSelected && strcmp(vItem.GetName, vCells.GetName)
            vCellsSelection = vNumberOfCells;
        end
    end
end

if min(vNumberOfSpots,vNumberOfCells) == 0
    msgbox('Please create some spots and a cell object')
    return
end

if vNumberOfSpots > 1
    [vSpotsSelection,vOk] = listdlg('ListString',vSpotsName, ...
        'InitialValue', vSpotsSelection, 'SelectionMode','single', ...
        'ListSize',[300 300], 'Name','Find Spots Close To Surface', ...
        'PromptString',{'Please Select one Spot Object:'});
    if vOk<1, return, end
end
if vNumberOfCells > 1 && ~vCellSelected
    [vCellsSelection,vOk] = listdlg('ListString',vCellsName, ...
        'InitialValue', vCellsSelection, 'SelectionMode','single', ...
        'ListSize',[300 300], 'Name','Find Spots Close To Cells', ...
        'PromptString',{'Please select one Cell Object:'});
    if vOk<1, return, end
end

% fill the cells to the image
vProgressBar = waitbar(0,'Vesicle Outside Cell');

vCellComponent = vImarisApplication.GetFactory.ToCells(vParent.GetChild(...
        vCellsList(vCellsSelection(1)) - 1));
 
vSpotComponent = vImarisApplication.GetFactory.ToSpots(vParent.GetChild(...
        vSpotsList(vSpotsSelection(1)) - 1));
  
vNumberOfCells = vCellComponent.GetNumberOfCells;
vTimeIndices = vSpotComponent.GetIndicesT;
vNofSpots = size(vTimeIndices,1);
if vNumberOfCells == 0 || vNofSpots == 0
  close(vProgressBar);
  return;
end
   
vCellIndicesT = vCellComponent.GetIndicesT;
vTimePointsWithCells = unique(vCellIndicesT);
vNofCellTimeIndices = numel(vTimePointsWithCells);

% find the closest cell for every spot
vSpotPositions = vSpotComponent.GetPositionsXYZ;
vSpotIndicesT = vSpotComponent.GetIndicesT;
vNofSpots = size(vSpotPositions,1);
vNearestCellIndexOfSpot = zeros(vNofSpots,1);
vDistanceToNearestCellOfSpot = zeros(vNofSpots,1);
vWaitbarValue = 0;
for vTimeIndex = 1:vNofCellTimeIndices
  waitbar(vWaitbarValue, vProgressBar, [ 'Vesicle Outside Cell, time point ' num2str(vTimeIndex) ' of ' num2str(vNofCellTimeIndices)]);  
  vTime = vTimePointsWithCells(vTimeIndex);
  vValidCellIndices = find(vCellIndicesT == vTime);
  % continue if there are no cells in current timepoint
  if numel(vValidCellIndices) == 0
    continue;
  end
  vCurrentCellID = vValidCellIndices(1);
  % Get the mask DataSet
  vNumberOfCells = numel(vValidCellIndices);
  vCell = vCellComponent.GetCell(vCurrentCellID-1);
  vMask1D = vCell.GetDataVolumeAs1DArrayFloats(0,0) * vCurrentCellID;
  % cell extend is the same for all cells
  vCellExtendMin = [vCell.GetExtendMinX vCell.GetExtendMinY vCell.GetExtendMinZ];
  vCellExtendMax = [vCell.GetExtendMaxX vCell.GetExtendMaxY vCell.GetExtendMaxZ];
  vCellSize = [vCell.GetSizeX vCell.GetSizeY vCell.GetSizeZ];
  vCellExtendDelta = (vCellExtendMax - vCellExtendMin) ./ vCellSize;
  for vCellIndex = 2:vNumberOfCells;
    vCurrentCellID = vValidCellIndices(vCellIndex);
    vCell = vCellComponent.GetCell(vCurrentCellID-1);
    vCellSize = [vCell.GetSizeX vCell.GetSizeY vCell.GetSizeZ];
    vCellExtendDelta = (vCellExtendMax - vCellExtendMin) ./ vCellSize;
    vMask1D = vMask1D + vCell.GetDataVolumeAs1DArrayFloats(0,0) * vCurrentCellID;
  end
  
  % this variable stores the cell ids + 1 (because the first cell ID in Imaris is 0
  vMaskVolume = reshape(vMask1D, vCell.GetSizeX, vCell.GetSizeY, vCell.GetSizeZ);

  % calculate the distance transform image
  [vDistanceImage, vNearestVoxelIndex] = bwdist(vMaskVolume);

  vCellIndexOfSpot = zeros(3,1) - 1;
  vValidSpotIndices = find(vSpotIndicesT == vTime);

  %vProgressBar = waitbar(0,'Getting Statistics ...');
  for i = 1:numel(vValidSpotIndices)
    vSpotIndex = vValidSpotIndices(i);
    vCurrentSpotPosition = vSpotPositions(vSpotIndex,:);
      if all(vCurrentSpotPosition >= vCellExtendMin) && all(vCurrentSpotPosition <= vCellExtendMax)
        vCellIndexOfSpot = floor((vCurrentSpotPosition - vCellExtendMin) ./ vCellExtendDelta);
      end

    % don't continue if any dimension is outside bounds
    if any(vCellIndexOfSpot == -1)
      vNearestCellIndexOfSpot(vSpotIndex) = -1;
      vDistanceToNearestCellOfSpot(vSpotIndex) = -1;
      continue
    end
    vCellIndexOfSpotMatlab = vCellIndexOfSpot + 1;
    vNearestVoxelIndexForSpot = vNearestVoxelIndex(vCellIndexOfSpotMatlab(1), vCellIndexOfSpotMatlab(2), vCellIndexOfSpotMatlab(3));
    vNearestCellIndexOfSpot(vSpotIndex) = vMaskVolume(vNearestVoxelIndexForSpot)-1;
    vDistanceToNearestCellOfSpot(vSpotIndex) = vDistanceImage(vCellIndexOfSpotMatlab(1), vCellIndexOfSpotMatlab(2), vCellIndexOfSpotMatlab(3));
  end
  vWaitbarValue = vTimeIndex/(vNofCellTimeIndices);
end

waitbar(vWaitbarValue, vProgressBar, 'Adding Statistics Values');  
%Create new statistics and add them
vNewNames = cell(vNofSpots,1);
vNewNames(:) = {'Nearest Cell ID'};
vNewValues = vNearestCellIndexOfSpot;
vNewUnits = cell(vNofSpots,1);
vNewUnits(:) = {''};
vNewFactors = cell(2,vNofSpots);
vNewFactors(1,:) = {'Spot'}; 
vNewFactors(2,:) = cellstr(num2str(vSpotIndicesT+1));
vNewFactorNames = {'Category'; 'Time'};
vNewIds = (0:vNofSpots-1)';
vSpotComponent.AddStatistics(vNewNames, vNewValues, vNewUnits, vNewFactors, vNewFactorNames, vNewIds);

vNewNames = cell(vNofSpots,1);
vNewNames(:) = {'Nearest Cell Distance'};
vNewValues = vDistanceToNearestCellOfSpot;
vNewUnits = cell(vNofSpots,1);
vDistanceUnit = char(vCellComponent.GetCell(0).GetUnit);
vNewUnits(:) = {vDistanceUnit};
vNewFactors = cell(2,vNofSpots);
vNewFactors(1,:) = {'Spot'}; 
vNewFactors(2,:) = cellstr(num2str(vSpotIndicesT+1));
vNewFactorNames = {'Category'; 'Time'};
vNewIds = (0:vNofSpots-1)';
vSpotComponent.AddStatistics(vNewNames, vNewValues, vNewUnits, vNewFactors, vNewFactorNames, vNewIds);
close(vProgressBar);


